; ASCII.ASM
; ZX Spectrum Text I/O routines - J.G.Harston
; -------------------------------------------
; The ZX Spectrum ROM does not have any general-purpose entry
; points for system calls. These routines implement some typical
; text I/O calls that are generally found on other systems. This
; library translates Spectrum characters to ASCII characters.
; Specifically:
;  DELETE key returns 127   WRCH 127 deletes previous character
;  BREAK  key returns 27    WRCH 8-13,30 moves cursor
;
; IOINIT  - Initialise I/O by selecting screen for output.
; IOTEST  - Test for pending keypress.
; IORDCH  - Wait for keyboard character, allowing symbol characters.
; IOWRCH  - Write a character to screen, allowing CHR$8-13,30.
;
; Requires  IY=>system variables, interupts on, IRQ routine at #38 running.


; Options
;IOCURSR EQU	143		; Cursor character
;IOCLICK EQU	0		; Key click
IOCURSR  EQU	1		; Cursor character, 1=show CAPS, 0=off
IOCLICK  EQU	1		; Key click, 0=no click
IOBREAK  EQU	1		; BREAK tested for
IOBUGFIX EQU	0		; Work around ZX ROM CHR$8 bugs


; Exported values:
IOCHDEL  EQU	127		; CHR$127 is delete output character.
IOKYDEL  EQU	127		; CHR$127 is DELETE input key.
IOKYBS   EQU	0		; No BACKSPC key.
IOCRLF   EQU	0		; No automatic LF after CR.


; IOINIT - Initialise screen driver
; ---------------------------------
; On exit,  AF,BC,DE,HL corrupted
IOINIT	LD	(IY+49),0	; Use whole screen
	LD	A,2
	JP	&1601		; Select screen and set up colours

; IOTEST - Test for input character
; ---------------------------------
; On exit, Z =no keypress pending
;          NZ=keypress pending
IOTEST	BIT	5,(IY+1)
	RET

; IORDCH - Wait for input character
; ---------------------------------
; On exit,  A =character
;           NC=character returned
;           CS=BREAK pressed, A=undefined
;           Alternative registers corrupted
IOCAPS	LD	A,(23658)
	XOR	8		; Toggle CAPSLOCK
	LD	(23658),A
IORDCH
  IF IOCURSR>31
	LD	A,IOCURSR
	RST	16		; Output cursor
  ENDIF
  IF IOCURSR=1
	SET	7,(IY+85)	; Flashing
	LD	A,&43
	BIT	3,(IY+48)	; Test CAPS
	JR	NZ,IORDNXT
	LD	A,&4C
IORDNXT	RST	16		; Output cursor
	RES	7,(IY+85)
  ENDIF
  IF IOCURSR
	LD	A,8
    IF IOBUGFIX
  	CALL	IOWRCH
    ELSE
	RET	16		; Back over cursor
    ENDIF
  ENDIF
IORDLP	HALT			; Wait for interupt
  IF IOBREAK
	CALL	&1F54		; Test BREAK key
	LD	A,27
	CCF
	JR	C,IORDOK	; BREAK pressed, with A=27
  ENDIF
	BIT	5,(IY+1)
	JR	Z,IORDLP	; Loop until key pressed
	LD	A,(23560)	; Get keypress
	RES	5,(IY+1)	; Set 'no key pressed'
	CP	6
	JR	Z,IOCAPS	; CAPS pressed
  IF IOCLICK
	PUSH	AF
	PUSH	BC
	PUSH	HL
	PUSH	IX
	LD	D,0
	LD	E,(IY-1)
	LD	HL,&00C8
	CALL	&03B5		; Keyclick
	POP	IX
	POP	HL
	POP	BC
	POP	AF
  ENDIF
	CP	12
	JR	Z,IOSYMB	; DELETE -> A=127
	AND	A		; NC, P=char<128
	JP	P,IORDOK	; Character pressed
	SUB	194
	JR	NC,IOSYMB
	LD	A,2
IOSYMB	AND	15		; Convert symbol key
	ADD	A,IOCONV & 255	; Index into conversion table
	PUSH	HL
	LD	L,A
	LD	H,IOCONV / 256
	LD	A,(HL)
	POP	HL
IORDOK
  IF IOCURSR
	PUSH	AF		; Save character and BREAK/OK flag
	LD	A,127
	CALL	IOWRCH		; Remove cursor
	POP	AF
  ENDIF
	RET
; *NB* Table must not wrap over page boundary
IOCONV	DEFB	126,124,0,93,91,0,0	; ~ | . ] [ . .
	DEFB	0,0,125,123,92,127	; . . } { \ DEL

; IOWRCH - Write character
; ------------------------
; On entry, A =character
; On exit,  All preserved
;           Alternative registers corrupted
IOWRCH	PUSH	AF		; Save everyhting so that
	PUSH	BC		; routines can be just
	PUSH	DE		; jumped to.
	PUSH	HL
	CALL	IOWROUT
	POP	HL
	POP	DE
	POP	BC
	POP	AF
	RET
IOWROUT	LD	(IY+82),255	; Prevent 'Scroll?'
	LD	B,32		; Prepare 32*CHAR and ' '
	CP	127
	JR	Z,IOWRDEL	; CHR$127 -> DELETE
	CP	B
	JP	NC,16		; CHR$>=32 - > printout
	CP	30
	JP	Z,&0DD6		; HOME
	CP	14
	RET	NC
	CP	12
	JP	Z,&0DAF		; CLS
	JR	NC,IOWR13	; CR
IOWRBS	CP	10
	JP	Z,IOWR10	; DOWN
  IF IOBUGFIX
	LD	BC,(23688)	; BC=POSN
  ENDIF
  	JR	NC,IOWR11	; UP
	CP	8
	JP	Z,IOWR08	; LEFT
	RET	C		; RIGHT
IOWR09	LD	B,1		; Cursor right by 1*SPC
IOWR10	LD	A,' '		; Cursor down by 32*SPC
	JR	IOWRSET
  IF IOBUGFIX
IOWR08	LD	A,B		; Work around the ZX ROM
	ADD	C		; CHR$8 bug
	CP	57
	RET	NC		; (0,0) - don't move
	INC	C		; Move posn left
	LD	A,C
	CP	&22
	JR	C,IOWRPSN	; Within screen, store it
	LD	C,&02
IOWR11	INC	B		; Move posn up
	LD	A,B
	CP	&19
	JR	C,IOWRPSN	; Within screen, store it
	DEC	B		; Move back into screen
IOWRPSN	JP	&0DEE		; CLSET1
  ELSE
IOWR08	LD	B,1		; Cursor left by 1*BS
IOWR11	LD	A,8		; Cursor up by 32*BS
 ENDIF
IOWRSET	SET	0,(IY+87)	; Set OVER
	LD	(IY+86),255	; Set MASKT
IOWRLP	PUSH	AF
	RST	16
	POP	AF
	DJNZ	IOWRLP
	RES	0,(IY+87)	; Reset OVER
	LD	(IY+86),0	; Reset MASKT
	RET
IOWR13	LD	A,(23684)
	AND	&E0		; Force DFCC to column 0
	LD	(23684),A
	LD	(IY+78),&21	; Force POSN to column 0
	RET
IOWRDEL
  IF IOBUGFIX
	LD	A,8
	CALL	IOWRBS
	LD	A,32
	RST	16
	LD	A,8
	JR	IOWRBS
  ELSE
	LD	A,8
	RST	16
	LD	A,B
	RST	16
	LD	A,8
	RST	16
	RET
  ENDIF
